/*
 * Multicast Testing code
 *
 * (C) 2009 Linux Foundation
 *		Christoph Lameter <cl@linux-foundation.org>
 */
#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdarg.h>
#include <getopt.h>
#include <pthread.h>
#include <linux/types.h>
#include <math.h>
#include <errno.h>
#include <ifaddrs.h>

#ifndef offsetof
#define offsetof(a,b) __builtin_offsetof(a,b)
#endif

#define barrier() __asm__ __volatile__("": : :"memory")

int verbose = 0;
int multicast_loop = 0;
int port = 9002;	/* Port to listen to or send to */

/*
 * Generic funcxtions
 */
void fatal(const char *x, ...)
{
	va_list ap;

	va_start(ap, x);
	vfprintf(stderr, x, ap);
	va_end(ap);
	exit(EXIT_FAILURE);
}

static inline long minimum(int a, int b)
{
	if (a < b)
		return a;
	return b;
}

static inline __u64 get_timestamp(void)
{
	struct timespec spec;

	clock_gettime(CLOCK_REALTIME, &spec);
	return spec.tv_nsec + spec.tv_sec * 1000000000LL;
}

struct in_addr toaddr(unsigned x)
{
	struct in_addr y = { x };

	return y;
}

char *ntoa(unsigned x)
{
	return inet_ntoa(toaddr(x));
}

unsigned aton(char *s)
{
	unsigned host;
	if (!inet_aton(s, (struct in_addr *)&host))
		fatal("Invalid address %s\n", s);
	return host;
}

void set_ip(struct sockaddr_in *s, unsigned host)
{
	s->sin_addr.s_addr = host;
	s->sin_family = AF_INET;
	s->sin_port = htons(port);
}

/*
 * Multicast patterns to use with the tests
 */
#define PATTERNS 10

struct highlow {
	unsigned char low;
	unsigned char high;
};

struct pattern {
	char *name;
	struct highlow nibble[4];
	unsigned max_groups;
	unsigned mlids;
	char *comments;
} pattern [PATTERNS] = {

  /* Standard patterns one MLID for each MC address */
  { "base239.2", { {239.239}, {2,2}, { 0, 255}, { 0, 255} }, 65536, 65536, "239.2.0.0 - 239.2.255.255" },
  { "base239.4", { {239.239}, {4,4}, { 0, 255}, { 0, 255} }, 65536, 65536, "239.4.0.0 - 239.4.255.255" },
  { "base239.100", { {239.239}, {100,100}, { 0, 255}, { 0, 255} }, 65536, 65536, "239.100.0.0 - 239.100.255.255" },
  { "base239.200",{ {239.239}, {200,200}, { 0, 255}, { 0, 255} }, 65536, 65536, "239.200.0.0 - 239.200.255.255" },
  /* Patterns that do not use 0 or 255 in any nibble */
  { "clean239.5",{ {239.239}, {5,5}, { 1, 254}, { 1, 254} }, 64516, 64516, "239.5.X.X no 0 and 255 nibbles" },
  { "clean239.6",{ {239.239}, {6,6}, { 1, 254}, { 1, 254} }, 64516, 64516, "239.6.X.X no 0 and 255 nibbles" },
  /* Voltaire supported IPoIB patterns to reduce MLID use */
  { "IPoIB1", { {224,224}, {1,10}, {10,10}, {1,255} }, 2550, 255, "Voltaire Pattern 1" },
  { "IPoIB2", { {230,234}, {1,1} , {1,2}  , {1,255} }, 2550, 510, "Voltaire Pattern 2" },
  { "IPoIB3", { {230,234}, {1,1} , {1,255}, {1,2}   }, 2550, 8, "Voltaire Pattern 3" },
  { "IPoIB4", { {224,233}, {1,1} , {1,3}  , {1,255} }, 7650, 765, "Voltaire Pattern 4" },
};

struct pattern *current_pattern = &pattern[0];
int offset_into_pattern = 0;
unsigned *pattern_ips;

/* Create the list of valid MC addresses for patterns */
unsigned *build_pattern_array(struct pattern *p)
{
	unsigned nibble[4];
	int i;
	int n;
	unsigned *pa;

	if (!p)
		return NULL;

	pa = malloc(sizeof(unsigned) * p->max_groups);
	for(i=0; i<4;i++)
		nibble[i] = p->nibble[i].low;

	n = 0;
	do {
		pa[n++] = htonl((nibble[0] << 24) + (nibble[1] << 16) + (nibble[2] << 8) + nibble[3]);

		i = 3;
		while (i >= 0) {
			nibble[i]++;
			if (nibble[i] <= p->nibble[i].high)
				break;
			nibble[i] = p->nibble[i].low;
			i--;
		}
	} while (i >= 0);

	if (n != p->max_groups)
		abort();
	return pa;
}

void print_patterns(void)
{
	int i;

	printf("\nPattern       MCadrs MLIDS Comments\n");
	printf("---------------------------------------------\n");
	for(i = 0; i < PATTERNS; i++) {
		printf("%12s %6d %6d %s\n", pattern[i].name, pattern[i].max_groups, pattern[i].mlids, pattern[i].comments);
	}
}

int parse_pattern(char *s)
{
	int i;

	for(i = 0; i < PATTERNS; i++)
		if (strcasecmp(s, pattern[i].name) == 0) {
			current_pattern = &pattern[i];
			return 0;
		}
	printf("Unknown pattern '%s'\n",s);
	print_patterns();
	return 1;
}

/*
 * Calculate the Nth IP address starting at a base ip address.
 */
unsigned inline multicast_group(int n)
{
	return pattern_ips[n + offset_into_pattern];
}

/*
 * Open socket
 */
int open_sock(int bind_socket, unsigned in, int ttl)
{
	int sock;
	int err;
	int yes = 1;
	int no = 0;

	sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (sock < 0) {
		perror("socket");
		fatal("socket error");
	}

	if (bind_socket) {
		struct sockaddr_in inad;

		err = setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes));
		if (err) {
			perror("reuseaddr");
			fatal("reuseaddr error");
		}


		set_ip(&inad, in);
		if (bind(sock, &inad, sizeof(inad))) {
			perror("bind");
			fatal("bind error");
		}
	}

	if (ttl >= 0) {
		err = setsockopt(sock, IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof(ttl));
		if (err) {
			perror("setsockopt IP_MULTICAST_TTL\n");
			fatal("setsockopt error");
		}
	}

#ifdef IP_MULTICAST_ALL
	setsockopt(sock, IPPROTO_IP, IP_MULTICAST_ALL, &no, sizeof(no));
#endif

	/*
	 * We are using two sockets. If we are not switching off loopback then everything
	 * we send will come back to us.
	 */
	if (!multicast_loop) {
		err = setsockopt(sock, IPPROTO_IP, IP_MULTICAST_LOOP, &no, sizeof(no));
		if (err) {
			perror("setsockopt IP_MULTICAST_LOOP\n");
			fatal("setsockopt error");
		}
	}

	return sock;
}

/*
 * Join a multicast group
 */
int mcast_join(int sock, unsigned group)
{

	struct ip_mreq mreq;
	struct sockaddr_in addr;

	set_ip(&addr, group);

	mreq.imr_multiaddr.s_addr = group;
	mreq.imr_interface.s_addr = htonl(INADDR_ANY);
	if (verbose > 1)
		printf("Joining mcast group %s\n", ntoa(group));

	if (setsockopt(sock, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq)) < 0) {
		perror("setsockopt(mcast join)");
		fatal("Join error");
	}

	return 0;
}

int mcast_leave(int sock, unsigned group)
{
	struct ip_mreq mreq;
	struct sockaddr_in addr;

	set_ip(&addr, group);

	mreq.imr_multiaddr.s_addr = group;
	mreq.imr_interface.s_addr = htonl(INADDR_ANY);
	if (verbose > 1)
		printf("Leaving mcast group %s\n", ntoa(group));

	if (setsockopt(sock, IPPROTO_IP, IP_DROP_MEMBERSHIP, &mreq, sizeof(mreq)) < 0) {
		perror("setsockopt(mcast_leave)");
		fatal("MCast Leave error");
	}

	return 0;
}

enum control { C_NOOP, C_DATA, C_TESTEND, C_NEW_MC_GROUP, C_RM_MC_GROUP,
	C_DUMP_STATISTICS, C_STARTUP, C_LAT_REQUEST, C_LAT_RESPONSE };

char *control_text[] = {
	"No Operation",
	"Data",
	"End of test",
	"New Multicast Group",
	"Remove Multicast Group",
	"Dump statistics",
	"Startup",
	"Latency Data request",
	"Latency Data response"
};

#define MAGIC 0xFEEDABCD

struct message {
	unsigned magic;
	unsigned to;		/* Sender */
	unsigned from;		/* MC or Unicast target */
	unsigned a;		/* Multicast group this is regarding or seqid */
	unsigned long long b;	/* Generic parameter. Usually timestamp */
	enum control control;	/* Control token */
	char payload[];
};

int ttl = 20;		/* Set ttl */
int length = 300;
int histogram = 0;
int mc_group_display = 0;
int suppress_errors = 0;
int terminate_remote = 0; /* Terminate remote listener after test */
__u64 lat_interval = 1000000000LL; /* latency interval 1s */
__u64 min_lat_interval = 1000000L; /* Min time between latency probes */
__u64 last_lat;			/* Last Latency probe */

int number_of_multicast_groups = 100;

unsigned bindaddr;		/* IP address to bind to */
unsigned control_channel;	/* Multicast control channel. Defaults to 239.3.3.3 */
unsigned localaddr;

int rate = 10;					/* Limit rate */
struct timespec wait_time = { 0, 0 };		/* Pause between packets */
unsigned long long receiver_start;
unsigned long long sender_start;
unsigned long long sample_start;
unsigned long long stat_interval = 10000000000LL;	/* 10 seconds */

int corrupted = 0;

/*
 * Format a message of length bytes
 */
void init_message(struct message *m, int length)
{
	int i;

	m->magic = MAGIC;
	m->from = localaddr;

	for (i = 0; i < length - offsetof(struct message, payload); i++)
		m->payload[i] = i;
}

/* Send UDP message to a target address with a control token and a sequence id */
void send_data_message(int sock, struct message *m, unsigned to, enum control c, unsigned seqid)
{
	struct sockaddr_in addr;

	if (verbose > 3)
		 printf("Sending message to %s, control=%s seqid=%d\n",
			ntoa(to), control_text[c], seqid);

	set_ip(&addr, to);

	m->to = to;
	m->control = c;
	m->a = seqid;
	m->b = get_timestamp();
	if (sendto(sock, m, length, 0, &addr, sizeof(addr)) < 0) {
		perror("sendto");
		fatal("Socket Send error");
	}
}

/* Send UDP message to a target address without a payload */
void send_message(int sock, struct message *m, unsigned to, enum control c,
		unsigned a, unsigned long long b)
{
	struct sockaddr_in addr;

	if (verbose)
		 printf("Sending short message to %s, control=%s from=%s\n",
			strdupa(ntoa(to)), control_text[c],strdupa(ntoa(m->from)));

	set_ip(&addr, to);

	m->to = to;
	m->control = c;
	m->a = a;
	m->b = b;
	if (sendto(sock, m, offsetof(struct message, payload), 0, &addr, sizeof(addr)) < 0) {
		perror("sendto");
		fatal("Socket short Send error");
	}

}

/*
 * Send a control message for all multicast groups that we send on
 */
void mc_group_notify(int sock, enum control c)
{
	struct message *m = alloca(length);

	init_message(m, length);
	send_message(sock, m, control_channel, c,
		multicast_group(0),
		(((unsigned long long)(current_pattern - pattern)) << 40) +
		((unsigned long long)number_of_multicast_groups << 20) +
		offset_into_pattern);
}

/*
 * Statistics subsystem
 */
#define MAX_US 100000L

int mc_channels;

struct statistics {
	unsigned mc_addr;
	unsigned host;
	unsigned seqid;
	struct statistics *next;
	unsigned long min,max;
	unsigned long long sum, sqsum;
	double avg;
	unsigned outliers, sequence_errors;
	unsigned long bytes;
	unsigned long messages;
	unsigned long samples;
	unsigned long long last_lat, start_time;
	unsigned long messages_last;
	unsigned long long timestamp_last;
	char *channel;		/* Name for pretty printing */
	unsigned histogram[MAX_US];
};

#define STAT_HASH_SIZE 1024

struct statistics *stat[STAT_HASH_SIZE];

int stathash(unsigned mc_addr)
{
	return mc_addr % STAT_HASH_SIZE;
}

struct statistics *__find_data(unsigned mc_addr)
{
	unsigned hash = stathash(mc_addr);
	struct statistics *s = stat[hash];

	while (s && s->mc_addr != mc_addr) {
		s = s->next;
	}
	return s;
}

void mc_join(int sock, unsigned group, int host)
{
	if (!__find_data(group)) {
		struct statistics *new = calloc(1, sizeof(struct statistics));
		int hash;

		new->mc_addr = group;
		new->host = host;
		new->seqid = 0;
		new->start_time = get_timestamp();
		new->channel = strdup(ntoa(group));
		mc_channels++;

		hash = stathash(group);
		new->next = stat[hash];
		stat[hash] = new;
		mcast_join(sock, group);
	}
}

void mc_leave(int sock, unsigned group)
{
	int hash = stathash(group);
	struct statistics *s = stat[hash];
	struct statistics *p = NULL;

	while (s && s->mc_addr != group) {
		p = s;
		s = s->next;
	}

	if (!s) {
		printf("Cannot find MC addr %s\n", ntoa(group));
		return;
	}

	if (p)
		p->next = s->next;
	else
		stat[hash] = s->next;

	mcast_leave(sock, group);
	free(s->channel);
	free(s);
}

struct statistics *find_data(unsigned mc_addr)
{
	struct statistics *s = __find_data(mc_addr);

	if (!s)
		fatal("Cannot find MC addr %s\n", ntoa(mc_addr));

	return s;
}

int hist[MAX_US];

void histo(int samples)
{
	int i, count, scale, max;

	printf("\nMicrosec\tPercent\tSamples\n");

	/* Conflate array to have less than 20 entries */
	count = 999999;
	scale = 1;
	max = MAX_US;

	count = 0;
	for (i = 0; i < max; i++)
		if (hist[i])
			count++;

	while (count > 20) {
		int newmax = max / 2;

		/* Conflate array */
		for (i = 0; i < newmax; i++)
			hist[i] = hist[ 2* i ]  + hist [ 2*i + 1 ];

		if (max % 2)
			hist[newmax - 1] += hist[max - 1];

		max = newmax;
		scale = scale * 2;
		/* Recount */
		count = 0;
		for (i = 0; i < max; i++)
			if (hist[i])
				count++;
	}

	for(i=0; i < max; i++)
		if (hist[i])
			printf("%4.0f-%.0f\t%4.0f%%\t%8u\n",
				(double)i * scale,
				(double)(i + 1) * scale,
				hist[i] * (double)100.0 / samples,
				hist[i]);
}

void print_statistics(void)
{
	unsigned i;
	struct statistics *s;
	int number_stats = 0;
	double total_kbytes = 0;
	double sstddev = 0;
	double savg = 0;
	unsigned long tmin = 0, tmax = 9;
	unsigned long total_sequence_errors = 0, total_messages = 0, total_outliers = 0, total_samples = 0;
	int idle =  0;
	unsigned long long ts = get_timestamp();
	double runtime = (ts - sender_start) / 1000000000.0;
	double sampletime = (ts - sample_start) / 1000000000.0;

	if (histogram)
		memset(hist, 0, sizeof(hist));

	if (mc_group_display) {
		/* Loop over all received channel and print statistics */
		printf("\n----------------Summary---t=%.0fsec-------------------------------------------------------------\n", runtime);
		printf("Messages Errors Msg/sec   Min/us  Avg/us  Max/us Stddev Outliers kbytes/sec   Group      Source    Samples\n");
		printf("----------------------------------------------------------------------------------------------------------\n");
	}

	for (i = 0 ; i < STAT_HASH_SIZE; i++)
		for (s = stat[i]; s; s = s->next) {

		unsigned long long run_time;
		unsigned messages_in_interval;
		unsigned long kbytes;
		double run_seconds;
		double avg, stddev;

		total_sequence_errors += s->sequence_errors;
		total_samples += s->samples;

		messages_in_interval = s->messages - s->messages_last;
		run_time = ts - s->timestamp_last;
		run_seconds = (double)run_time / (double)1000000000.0;
		if (!messages_in_interval)
			idle++;

		s->messages_last = s->messages;
		s->timestamp_last = ts;

		total_messages += messages_in_interval;
		kbytes = (s->bytes + 500) / 1000;
		total_kbytes += kbytes;

		if (s->samples) {
			avg = (double) s->sum / s->samples;
			savg += avg;
			if (s->max > tmax)
				tmax = s->max;
			if (s->min < tmin || !tmin)
				tmin = s->min;
			stddev = sqrt((double)s->sqsum / s->samples - avg * avg);
			sstddev += stddev;
			if (histogram) {
				int i;
				for(i = 0 ; i < MAX_US; i++)
					hist[i] += s->histogram[i];
			}
			number_stats++;
		}

		if (mc_group_display) {
			printf("%8ld %6d %7.0f ", s->messages, s->sequence_errors, messages_in_interval / run_seconds);
			printf("%7.2f ", s->min / 1000.0);
			printf("%7.2f ", avg);
			printf("%7.2f ", s->max / 1000.0);
			printf("%7.2f ", stddev);
			printf("%8d ", s->outliers);
			printf("%7.1f ",  kbytes / run_seconds);
			printf("%12s %12s %6ld\n", s->channel, inet_ntoa(toaddr(s->host)), s->samples);
		}
		s->sequence_errors = s->samples = s->bytes = s->sum = s->sqsum = s->min = s->max = 0;
	}

	if (mc_group_display)
		/* Sum it all up and print more statistics */
		printf("----Totals------------------------------Groups=%d-----------------------Corrupted-----Samples----\n",
				mc_channels);
	else {
		static int first = 1;

		if (first) {
			printf("\n   Time TotalMsg SeqErr Msg/Sec   Min/us  Avg/us  Max/us StdDev Outliers  Kbytes Idle Corrupted Samples\n");
			first = 0;
		}
		printf("%7.0f ", runtime);
	}
	printf("%8ld %6ld %7.0f ", total_messages, total_sequence_errors, total_messages / sampletime);
	printf("%7.2f ", tmin / 1000.0);
	printf("%7.2f ", savg / number_stats);
	printf("%7.2f ", tmax / 1000.0);
	printf("%7.2f ", sstddev / number_stats);
	printf("%8ld ", total_outliers);
	printf("%7.1f ", total_kbytes);
	printf("%4d ", idle);
	printf("%8d ", corrupted);
	printf("%8ld\n", total_samples);

	corrupted = 0;

	if (histogram)
		histo(total_samples);
	sample_start = get_timestamp();
}

/*
 * Sender thread
 *
 * Thread sends messages to all MC groups in a loop and then delays to get the proper timing.
 *
 * Statistics are generated in regular intervals. MC groups are announced when statistics are
 * generated.
 */
enum { SENDER_OFF, SENDER_SYNC, SENDER_RUNNING, SENDER_STOP_REQUEST } sender_mode;
enum { RECEIVER_OFF, RECEIVER_ON } receiver_mode;

void *sender_thread(void *dummy)
{
	int i;
	int seqid;
	int sock;
	struct message *m = alloca(length);
	unsigned long long ts;
	unsigned long long next_stats;

	sock = open_sock(0, 0, ttl);
	init_message(m, length);

	/*
	 * Start sending
	 */
	seqid = 1;

	sender_mode = SENDER_SYNC;

resync:
	while (receiver_mode != RECEIVER_ON)
			barrier();

	printf("Sender: Announcing %d multicast groups from %s pattern %s offset %d\n",
				number_of_multicast_groups,
				ntoa(multicast_group(0)),
				current_pattern->name,
				offset_into_pattern);
	mc_group_notify(sock, C_NEW_MC_GROUP);
	sleep(1);

	printf("Sender: Sending %d msgs/ch/sec. Latency probe interval=%g sec/ch. Min probe interval=%g sec\n",
		rate,
		lat_interval / 1000000000.0,
		min_lat_interval / 1000000000.0);

	sender_mode = SENDER_RUNNING;
	sender_start = get_timestamp();
	sample_start = sender_start;
	next_stats = sender_start + stat_interval;

	while (sender_mode == SENDER_RUNNING) {
		__u64 next;
		int wait_ns;

		ts = get_timestamp();
		next = ts + 1000000000LL / rate;

		for (i = 0; i < number_of_multicast_groups; i++)
			send_data_message(sock, m, multicast_group(i), C_DATA, seqid);

		seqid++;
		ts = get_timestamp();
		if (ts > next_stats) {
			print_statistics();
			next_stats = ts + stat_interval;

			/* Reannounce the multicast groups that we carry */
			mc_group_notify(sock, C_NEW_MC_GROUP);
			ts = get_timestamp();
		}

		wait_ns = next - ts;
		if (wait_ns > 500) {
			wait_time.tv_nsec = wait_ns;
                       	nanosleep(&wait_time, NULL);
		}
	}
	if (sender_mode == SENDER_SYNC)
		goto resync;

	/*
	 * Close sender
	 */
	mc_group_notify(sock, C_RM_MC_GROUP);
	close(sock);
	sender_mode = SENDER_OFF;
	return NULL;
}


void start_sender(void)
{
	pthread_attr_t attr;
	pthread_t sender_thread_t;
	int rc;

	sender_mode = SENDER_OFF;

	pattern_ips = build_pattern_array(current_pattern);
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	rc = pthread_create(&sender_thread_t, &attr, sender_thread, NULL);
        if (rc < 0)
		fatal("Cannot create sender thread\n");
	if (verbose)
		printf("Started sender\n");
}

void stop_sender(void)
{
	if (verbose)
		printf("Stopping sender\n");

	if (sender_mode == SENDER_OFF)
		return;

	while (sender_mode != SENDER_RUNNING)
		barrier();
	sender_mode = SENDER_STOP_REQUEST;
	while (sender_mode != SENDER_OFF)
		barrier();
}

/*
 * Receiver thread
 *
 * Simply waits for traffic to come in on a port and then processes the data.
 */

/* Special messaging structure for replies */
struct message *reply;

/* Receiver has requested latency information for a message and provided a timestamp */
void lat_request(int sock, struct message *m, unsigned long long ts)
{
	unsigned group = m->a;
	unsigned long long timestamp = m->b;

	if (verbose > 1)
		printf("Latency request from %s\n", ntoa(m->from));

	send_message(sock, reply, m->from, C_LAT_RESPONSE, group, (ts - timestamp) / 2);
}

void lat_response(int sock, struct message *m)
{
	unsigned group = m->a;
	unsigned long diff = m->b;
	struct statistics *s = find_data(group);

	if (verbose)
		printf("Latency response from %s diff %ld nsec\n", ntoa(m->from), diff);

	/* Record data */
	if (diff < 0 || (histogram && diff > MAX_US * 100)) {
		if (s->outliers < 10) {
			if (!suppress_errors) {
				printf("Outlier #%d: Latency=%7.2fusecs\n",
					s->outliers + 1, diff / 1000.0);
				if (s->outliers == 9)
					printf("Too many outliers.\n");
			}
		}
		s->outliers++;
	} else {
		int udiff;

		if (diff < s->min || s->min == 0)
			s->min = diff;
		if (diff > s->max)
			s->max = diff;

		udiff = diff / 1000;
		if (histogram)
			s->histogram[ udiff]++;
		s->sum += udiff;
		s->sqsum += udiff * udiff;
		s->samples++;
	}
}

void handle_data(int sock, struct message *m, __u64 ts, unsigned length)
{
	struct statistics *s = find_data(m->to);
	unsigned seqid = m->a;
	unsigned long long timestamp = m->b;

	if (seqid != s->seqid) {
		if (s->seqid && seqid != 1) {
			if (!suppress_errors)
				printf("*** Channel %s: Message out of sequence. Expected %d received %d length=%d\n",
					s->channel, s->seqid, seqid, length);
			s->sequence_errors++;
		}
		s->seqid = seqid;
	}

	if (verbose > 2)
		printf("Channel %s: %d Received data (from %s). length=%d\n",
			s->channel, seqid,
			strdupa(ntoa(m->from)),
			length);

	s->messages++;
	s->seqid++;
	s->bytes += length;
	if (s->last_lat + lat_interval < ts && last_lat + min_lat_interval < ts) {
		send_message(sock, reply, m->from, C_LAT_REQUEST, m->to, timestamp);
		s->last_lat = last_lat = ts;
	}
}

void new_mc_groups(int sock, struct message *m)
{
	unsigned group = m->a;
	struct pattern *p = &pattern[m->b >> 40];
	unsigned offset = m->b & ((1 << 20) - 1);
	unsigned groups = (m->b >> 20) & ((1 << 20) - 1);
	unsigned send_first_mc = offset_into_pattern;
	unsigned send_last_mc = offset_into_pattern + number_of_multicast_groups;
	unsigned first_mc = offset;
	unsigned last_mc = offset + groups - 1;

	if (p == current_pattern && !multicast_loop &&
		((first_mc >= send_first_mc && first_mc < send_last_mc) ||
		(last_mc  >= send_first_mc && last_mc < send_last_mc) ||
		(first_mc < send_first_mc && last_mc >= send_last_mc)))
			printf("Rejecting MC groups from %s due to overlap in pattern %s\n",
				ntoa(m->from), p->name);

	else if (!__find_data(group)) {
		int i;
		unsigned *ips = build_pattern_array(p);

		printf("Subscribing to %d new multicast addresses first %s pattern=%s offset %d origin %s\n",
			groups, strdupa(ntoa(group)), p->name, offset, strdupa(ntoa(m->from)));

		for (i = 0; i < groups; i++)
			mc_join(sock, ips [offset + i], m->from);

		free(ips);
	}
}

void remove_mc_groups(int sock, struct message *m)
{
	unsigned group = m->a;
	struct pattern *p = &pattern[m->b >> 40];
	unsigned offset = m->b & ((1 << 20) - 1);
	unsigned groups = (m->b >> 20) & ((1 << 20) - 1);

	if (__find_data(group)) {
		int i;
		unsigned *ips = build_pattern_array(p);

		for (i = 0; i < groups; i++)
			mc_leave(sock, ips [offset + i]);
		free(ips);
	}
}

void receiver_loop(void)
{
	struct message *m = alloca(length);
	int sock;
	unsigned long long ts;
	struct sockaddr_in inaddr;
	unsigned inaddrlen = sizeof(struct sockaddr_in);
	unsigned long total_messages = 0;

	/* initialize variables */
	corrupted = 0;
	sock = open_sock(1, bindaddr, ttl);

	set_ip(&inaddr, bindaddr);
	/* Join control channel */
	mcast_join(sock, control_channel);

	printf("Receiver: Listening to control channel %s\n", ntoa(control_channel));
	receiver_start = get_timestamp();
	receiver_mode = RECEIVER_ON;

	do {
		int nbytes;

		nbytes = recvfrom(sock, m, length, 0, &inaddr, &inaddrlen);
		ts = get_timestamp();
		if (nbytes < 0) {
			perror("recvfrom");
			break;
		}

		if (verbose > 2)
			printf("Receiver_loop: %d bytes cmd=%s from=%s to=%s\n",
				nbytes, control_text[m->control],strdupa(ntoa(m->from)), strdupa(ntoa(m->to)));

		if (m->magic != MAGIC) {
			printf("** Corrupted packet ignored MAGIC=%x to=%s from=%s\n",m->magic, strdupa(ntoa(m->to)), strdupa(ntoa(m->from)));
			corrupted++;
			continue;
		}

		switch (m->control) {

		case C_NOOP:
			/* Ignore */
			break;

		case C_DATA:
			handle_data(sock, m, ts, nbytes);
			break;

		case C_LAT_REQUEST:
			lat_request(sock, m, ts);
			break;

		case C_LAT_RESPONSE:
			lat_response(sock, m);
			break;

		case C_STARTUP:
			if (sender_mode == SENDER_RUNNING)
				sender_mode = SENDER_SYNC;
			break;

		case C_NEW_MC_GROUP:
			new_mc_groups(sock, m);
			break;

		case C_RM_MC_GROUP:
			remove_mc_groups(sock, m);
			break;

		case C_DUMP_STATISTICS:
			print_statistics();
			break;

		case C_TESTEND:
			goto out;

		}
		total_messages++;
	} while (1);
out:
	receiver_mode = RECEIVER_OFF;
	close(sock);
}


void find_local_addr(void)
{
	struct ifaddrs *ifaddr, *ifa;
	int family;
	int nr_match = 0;

	if (getifaddrs(&ifaddr) == -1 ) {
		perror("getifaddr");
		fatal("bad");
	}

	for (ifa = ifaddr; ifa; ifa = ifa->ifa_next) {
		if (!ifa->ifa_addr)
			continue;

		family = ifa->ifa_addr->sa_family;

		if (family != AF_INET)
			continue;

		/* If we are bound to an interface then the IP must match */
		if (bindaddr && ((struct sockaddr_in *)ifa->ifa_addr)->sin_addr.s_addr != bindaddr)
			continue;

		if (strcmp(ifa->ifa_name, "lo") == 0)
			continue;

		localaddr = ((struct sockaddr_in *)ifa->ifa_addr)->sin_addr.s_addr;
		nr_match++;
	}
	freeifaddrs(ifaddr);
	if (nr_match > 1)
		printf("WARNNIG: Multiple active ethernet devices. Using local address %s\n", ntoa(localaddr));
}

/*
 * Command line processing
 */
void usage(void)
{
	printf("mcasttorture: Multicast stress test 7/10/2009\n"
		"(C) 2009 Linux Foundation Christoph Lameter <cl@linux-foundation.org>\n\n"
		"mcasttest {<options>} <number of multicast channels to create>\n\n"
		"-b<ip>|--bind <ip>                Bind sockets to a local NICs IP address\n"
		"-c<ip>|--control <ip>             Multicast control channel [239.3.3.3]\n"
		"-d|--details                      Print histogram\n"
		"-e|--errors                       Suppress printing communication errors\n"
		"-h|--help                         Print usage information\n"
		"-i<sec>|--interval <sec>          Statistics display interval[1]\n"
		"-j<pattern>[+offset]              Use predefined pattern(see below)\n"
		"-l<length>|--length <length>      Set length of packet[300]\n"
		"-m|--multicast                    Display multicast groups in statistics\n"
		"-o|--offset                       mc group offset into pattern\n"
		"-p<port>|--port <port>            Change the port to use [9002]\n"
		"-r<packets>|--rate <packets>      Packets to send per second(per mc group)[10].\n"
		"-v|--verbose                      Increase detail of logging\n"
		"-X|--exit                         Terminate all mcasttorture instances\n"
		"-L|--loop                         Multicast loopback (required for localhost)\n"
	);
	print_patterns();
}

struct option opts[] = {
	{ "port", 1, NULL, 'p' },
	{ "number", 1, NULL, 'n' },
	{ "time", 1, NULL, 't' },
	{ "length", 1, NULL, 'l' },
	{ "bind", 1, NULL, 'b' },
	{ "help", 0, NULL, 'h' },
	{ "verbose", 0, NULL, 'v' },
	{ "rate", 1, NULL, 'r' },
	{ "ttl", 1, NULL, 't' },
	{ "join", 1, NULL, 'j' },
	{ "details", 1, NULL, 'd' },
	{ "terminaate", 0, NULL, 'X' },
	{ "loop", 0, NULL, 'L' },
	{ "offset", 1, NULL, 'o' },
	{ NULL, 0, NULL, 0 }
};

int main(int argc, char *argv[])
{
	int c;

	control_channel = aton("239.3.3.3");

	while ((c = getopt_long(argc, argv, "b:c:dehi:j:l:mo:p:r:t:vXL",
							opts, NULL)) != -1)
	switch (c) {
	case 'b' :
		bindaddr = aton(optarg);
		break;

	case 'c' :
		control_channel = aton(optarg);
		break;

	case 'd':
		histogram = 1;
		break;

	case 'e':
		suppress_errors = 1;
		break;

	case 'h' :
		usage();
		return 0;

	case 'i' :
		stat_interval = atof(optarg) * 1000000000LL;
		break;

	case 'j' :
		if (parse_pattern(optarg))
			fatal("Invalid pattern\n");
		break;

	case 'l' :
		length = atoi(optarg);
		if (length < offsetof(struct message, payload) || length > 65000)
			fatal("Invalid packet length %d\n", length);
		break;

	case 'm':
		mc_group_display = 1;
		break;

	case 'o':
		offset_into_pattern = atoi(optarg);
		break;

	case 'p' :
		port = atoi(optarg);
		if (port < 1024 || port > 60000)
			fatal("Invalid port number %d\n", port);
		break;

	case 'r' :
		rate = atoi(optarg);
		if (rate > 1000000 || rate < 1)
			fatal("Invalid rate %d\n", rate);

		wait_time.tv_nsec = 1000000000 / rate;
		break;

	case 't':
		ttl = atoi(optarg);
		break;

	case 'v':
		verbose++;
		break;

	case 'X':
		terminate_remote = 1;
		break;

	case 'L':
		multicast_loop = 1;
		break;

	default:
		fatal("%s: Invalid option '%c'\n", argv[0], optopt);
	}

	if (argc > 1 + optind) {
		usage();
		return 1;
	}

	find_local_addr();
	reply = calloc(1, length);
	init_message(reply, length);

	if (argc == 1 + optind) {
		number_of_multicast_groups = atoi(argv[optind]);
		if (number_of_multicast_groups < 1 || number_of_multicast_groups > 10000)
			fatal("Invalid number of multicast groups\n");
		if (offset_into_pattern + number_of_multicast_groups > current_pattern->max_groups)
			fatal("Pattern does not support that many MC addresses\n");

		start_sender();
		receiver_loop();
	} else {
		printf("Must specify number of multicast channels\n");
		usage();
		return 1;
	}
	stop_sender();
	return 0;
}

